searchNodes=[{"ref":"StateServer.html","title":"StateServer","type":"behaviour","doc":"A wrapper for :gen_statem which preserves GenServer-like semantics. Motivation The :gen_statem event callback is complex, with a confusing set of response definitions, the documentation isn&#39;t that great, the states of the state machine are a bit too loosey-goosey and not explicitly declared anywhere in a single referential place in the code; you have to read the result bodies of several function bodies to understand the state graph. StateServer changes that. There are three major objectives: Fanout the callback handling Unify callback return type with that of GenServer, and sanitize Enforce the use of a programmer-defined state graph. Defining the state graph The state graph is defined at compile time using the keyword list in the use statement. This state_graph is a keyword list of keyword lists. The outer keyword list has the state names (atoms) as keys and the inner keyword lists have transitions (atoms) as keys, and destination states as values. The first keyword in the state graph is the initial state of the state machine. Defining the state graph is required. At compile time, StateServer will verify that all of the state graph&#39;s transition destinations exist as declared states; you may need to explicitly declare that a particular state is terminal by having it key into the empty list []. Example the state graph for a light switch might look like this: use StateServer, on: [flip: :off], off: [flip: :on] &#39;Magic&#39; things The following guards will be defined for you automatically. is_terminal/1: true iff the argument is a terminal state. is_terminal/2: true iff starting from &amp;1 through transition &amp;2 leads to a terminal state. is_transition/2: true iff &amp;2 is a proper transition of &amp;1 is_transition/3: true iff starting from &amp;1 through transition &amp;2 leads to &amp;3 The following types are defined for you automatically. state which is a union type of all state atoms. transition which is a union type of all transition atoms. The following module attributes are available at compile-time: @state_graph is the state graph as passed in the use statement @initial_state is the initial state of the state graph. Note that there are cases when the StateServer itself should not start in that state, for example if it is being restarted by an OTP supervisor and should search for its state from some other source of ground truth. State machine data A StateServer, like all :gen_statems carry additional data of any term in addition to the state, to ensure that they can perform all Turing-computable operations. You are free to make the data parameter whatever you would like. It is encouraged to declare the data type in the module which defines the typespec of this state machine data. Callbacks The following callbacks are all optional and are how you implement functionality for your StateServer. External callbacks: handle_call/4 responds to a message sent via GenServer.call/3. Like GenServer.handle_call/3, the calling process will block until you a reply, using either the {:reply, reply} tuple, or, if you emit :noreply, a subsequent call to reply/2 in a continuation. Note that if you do not reply within the call&#39;s expected timeout, the calling process will crash. handle_cast/3 responds to a message sent via GenServer.cast/2. Like GenServer.handl_cast/2, the calling process will immediately return and this is effectively a fire and forget operation with no backpressure response. handle_info/3 responds to a message sent via send/2. Typically this should be used to trap system messages that result from a message source that has registered the active StateServer process as a message sink, such as network packets or :nodeup/:nodedown messages (among others). Internal callbacks handle_internal/3 responds to internal events which have been sent forward in time using the {:internal, payload} setting. This is :gen_statem&#39;s primary method of doing continuations. If you have code that you think will need to be compared against or migrate to a :gen_statem, you should use this semantic. handle_continue/3 responds to internal events which have been sent forward in time using the {:continue, payload} setting. This is GenServer&#39;s primary method of performing continuations. If you have code that you think will need to be compared against or migrate to a GenServer, you should use this form. A typical use of this callback is to handle a long-running task that needs to be triggered after initialization. Because start_link/2 will timeout, if StateMachine, then you should these tasks using the continue callback. handle_timeout/3 handles all timeout events. See the timeout section for more information handle_transition/3 is triggered whenever you change states using the {:transition, transition} event. Note that it&#39;s not triggered by a {:goto, state} event. You may find the c:is_edge/3 callback guard to be useful for discriminating which transitions you care about. Special callbacks on_state_entry/3 will be triggered for the starting state (whether as a default or as set by a goto: parameter in init/1), and when any event causes the state machine to change state or repeat state Callback responses handle_call/4 typically issues a reply response. A reply response takes the one of two forms, {:reply, reply} or {:reply, reply, event_list} It may also take the noreply form, with a deferred reply at some other time. all of the callback responses may issue a noreply response, which takes one of two forms, :noreply or {:noreply, event_list} The event list The event list consists of one of several forms: {:transition, transition} # sends the state machine through the transition {:update, new_data} # updates the data portion of the state machine {:goto, new_state} # changes the state machine state without a transition {:internal, payload} # sends an internal event {:continue, payload} # sends a continuation {:event_timeout, {payload, time}} # sends an event timeout with a payload {:event_timeout, time} # sends an event timeout without a payload {:state_timeout, {payload, time}} # sends a state timeout with a payload {:state_timeout, time} # sends a state timeout without a payload {:timeout, {name, payload, time}} # sends a plain timeout with a name, and a payload {:timeout, {name, time}} # sends a plain timeout with a name, but no payload {:timeout, time} # sends a plain timeout without a payload :noop # does nothing transition and update events are special. If they are at the head of the event list, (and in that order) they will be handled atomically in the current function call; if they are not at the head of the event list, separate internal events will be generated, and they will be executed as separate calls in their event order. Typically, these should be represented in the event response as part of an Elixir keyword list, for example: {:noreply, transition: :flip, internal: {:add, 3}, state_timeout: 250} You may also generally provide events as tuples that are expected by :gen_statem, for example: {:next_event, :internal, {:foo, &quot;bar&quot;}}, but note that if you do so Elixir&#39;s keyword sugar will not be supported. Transition vs. goto Transitions represent the main business logic of your state machine. They come with an optional transition handler, so that you can write code that will be ensured to run on all state transitions with the same name, instead of requiring these to be in the code body of your event. You should be using transitions everywhere. However, there are some cases when you will want to skip straight to a state without traversing the state graph. Here are some cases where you will want to do that: If you want to start at a state other than the head state, depending on environment at the start If you want to do a unit test and skip straight to some state that you&#39;re testing. If your gen_statem has crashed, and you need to restart it in a state that isn&#39;t the default initial state. Timeouts StateServer state machines respect three types of timeouts: :event_timeout. These are cancelled when any internal OR external event hits the genserver. Typically, an event_timeout definition should be the last term in the event list, otherwise the succeeding internal event will cancel the timeout. :state_timeout. These are cancelled when the state of the state machine changes. NB a state machine may only have one state timeout active at any given time. :timeout. These are not cancelled, unless you reset their value to :infinity. In general, if you need to name your timeouts, you should include the &quot;name&quot; of the timeout in the &quot;payload&quot; section, as the first element in a tuple; you will then be able to pattern match this in your handle_timeout/3 headers. If you do not include a payload, then they will be explicitly sent a nil value. Organizing your code If you would like to organize your implementations by state, consider using the StateServer.State behaviour pattern. Example basic implementation: defmodule Switch do @doc &quot;&quot;&quot; implements a light switch as a state server. In data, it keeps a count of how many times the state of the light switch has changed. &quot;&quot;&quot; use StateServer, off: [flip: :on], on: [flip: :off] @type data :: non_neg_integer def start_link, do: StateServer.start_link(__MODULE__, :ok) @impl true def init(:ok), do: {:ok, 0} ############################################################## ## API ENDPOINTS @doc &quot;&quot;&quot; returns the state of switch. &quot;&quot;&quot; @spec state(GenServer.server) :: state def state(srv), do: GenServer.call(srv, :state) @spec state_impl(state) :: StateServer.reply_response defp state_impl(state) do {:reply, state} end @doc &quot;&quot;&quot; returns the number of times the switch state has been changed, from either flip transitions or by setting the switch value &quot;&quot;&quot; @spec count(GenServer.server) :: non_neg_integer def count(srv), do: GenServer.call(srv, :count) @spec count_impl(non_neg_integer) :: StateServer.reply_response defp count_impl(count), do: {:reply, count} @doc &quot;&quot;&quot; triggers the flip transition. &quot;&quot;&quot; @spec flip(GenServer.server) :: state def flip(srv), do: GenServer.call(srv, :flip) @spec flip_impl(state, non_neg_integer) :: StateServer.reply_response defp flip_impl(:on, count) do {:reply, :off, transition: :flip, update: count + 1} end defp flip_impl(:off, count) do {:reply, :on, transition: :flip, update: count + 1} end @doc &quot;&quot;&quot; sets the state of the switch, without explicitly triggering the flip transition. Note the use of the builtin `t:state/0` type. &quot;&quot;&quot; @spec set(GenServer.server, state) :: :ok def set(srv, new_state), do: GenServer.call(srv, {:set, new_state}) @spec set_impl(state, state, data) :: StateServer.reply_response defp set_impl(state, state, _) do {:reply, state} end defp set_impl(state, new_state, count) do {:reply, state, goto: new_state, update: count + 1} end ####################################################3 ## callback routing @impl true def handle_call(:state, _from, state, _count) do state_impl(state) end def handle_call(:count, _from, _state, count) do count_impl(count) end def handle_call(:flip, _from, state, count) do flip_impl(state, count) end def handle_call({:set, new_state}, _from, state, count) do set_impl(state, new_state, count) end # if we are flipping on the switch, then turn it off after 300 ms # to conserve energy. @impl true def handle_transition(state, transition, _count) when is_transition(state, transition, :on) do {:noreply, state_timeout: {:conserve, 300}} end def handle_transition(_, _, _), do: :noreply @impl true def handle_timeout(:conserve, :on, _count) do {:noreply, transition: :flip} end end"},{"ref":"StateServer.html#call/3","title":"StateServer.call/3","type":"function","doc":"should be identical to GenServer.call/3 NB: this behavior is consistent with the GenServer call but NOT the :gen_statem.call/3, which spawns a proxy process. StateServer chooses the GenServer call to maintain consistency across developer expectations. If you need :gen_statem-like behavior, you can manually call :gen_statem.call/3 passing the pid or reference and it should work as expected."},{"ref":"StateServer.html#cast/2","title":"StateServer.cast/2","type":"function","doc":"should be identical to GenServer.cast/2"},{"ref":"StateServer.html#defer/1","title":"StateServer.defer/1","type":"macro","doc":"a shortcut which lets you trap all other cases and send them to be handled by individual state modules. defer handle_call is equivalent to def handle_call(_, _, _, _), do: :defer"},{"ref":"StateServer.html#defstate/2","title":"StateServer.defstate/2","type":"macro","doc":"Like defstate/3 but lets you define your module externally. Example: defstate Some.Other.Module, for: :on"},{"ref":"StateServer.html#defstate/3","title":"StateServer.defstate/3","type":"macro","doc":"Defines a state module to organize your code internally. Keep in mind that the arities of all of the callbacks are less one since the associated state is bound in the parent module. Example: defstate On, for: :on do @impl true def handle_call(_, _, _) do ... end end"},{"ref":"StateServer.html#c:handle_call/4","title":"StateServer.handle_call/4","type":"callback","doc":"handles messages sent to the StateMachine using StateServer.call/3"},{"ref":"StateServer.html#c:handle_cast/3","title":"StateServer.handle_cast/3","type":"callback","doc":"handles messages sent to the StateMachine using StateServer.cast/2"},{"ref":"StateServer.html#c:handle_continue/3","title":"StateServer.handle_continue/3","type":"callback","doc":"handles events sent by the {:continue, payload} event response. NB a continuation is simply an :internal event with a reserved word tag attached."},{"ref":"StateServer.html#c:handle_info/3","title":"StateServer.handle_info/3","type":"callback","doc":"handles messages sent by send/2 or other system message generators."},{"ref":"StateServer.html#c:handle_internal/3","title":"StateServer.handle_internal/3","type":"callback","doc":"handles events sent by the {:internal, payload} event response."},{"ref":"StateServer.html#c:handle_timeout/3","title":"StateServer.handle_timeout/3","type":"callback","doc":"triggered when a set timeout event has timed out. See timeouts"},{"ref":"StateServer.html#c:handle_transition/3","title":"StateServer.handle_transition/3","type":"callback","doc":"triggered when a state change has been initiated via a {:transition, transition} event. should emit :noreply, or {:noreply, extra_actions} to handle the normal case when the transition should proceed. If the transition should be cancelled, emit :cancel or {:cancel, extra_actions}. NB: you may want to use the is_terminal/2 or the is_transition/3 callback defguards here."},{"ref":"StateServer.html#c:init/1","title":"StateServer.init/1","type":"callback","doc":"starts the state machine, similar to GenServer.init/1 NB the expected response of init/1 is {:ok, data} which does not include the initial state. The initial state is set as the first key in the :state_graph parameter of the use StateServer directive. If you must initialize the state to something else, use the {:ok, data, goto: state} response. You may also respond with the usual GenServer.init/1 responses, such as: :ignore {:stop, reason} You can also initialize and instrument one of several keyword parameters. For example, you may issue {:internal, term} or {:continue, term} to send an internal message as part of a startup continuation. You may send {:timeout, {term, timeout}} to send a delayed continuation; this is particularly useful to kick off a message loop. Any of these keywords may be preceded by {:goto, state} which will set the initial state, which is useful for resurrecting a supervised state machine into a state without a transition. Example def init(log) do # uses both the goto and the timeout directives to either initialize # a fresh state machine or resurrect from a log. In both cases, # sets up a ping loop to perform some task. case retrieve_log(log) do nil -&gt; {:ok, default_value, timeout: {:ping, 50}} {previous_state, value} -&gt; {:ok, value, goto: previous_state, timeout: {:ping, 50}} end end # for reference, what that ping loop might look like. def handle_timeout(:ping, _state, _data) do do_ping(:ping) {:noreply, timeout: {:ping, 50}} end"},{"ref":"StateServer.html#c:is_terminal/1","title":"StateServer.is_terminal/1","type":"macrocallback","doc":"an autogenerated guard which can be used to check if a state is terminal"},{"ref":"StateServer.html#c:is_terminal/2","title":"StateServer.is_terminal/2","type":"macrocallback","doc":"an autogenerated guard which can be used to check if a state and transition will lead to a terminal state."},{"ref":"StateServer.html#c:is_transition/2","title":"StateServer.is_transition/2","type":"macrocallback","doc":"an autogenerated guard which can be used to check if a transition is valid for a state"},{"ref":"StateServer.html#c:is_transition/3","title":"StateServer.is_transition/3","type":"macrocallback","doc":"an autogenerated guard which can be used to check if a state and transition will lead to any state."},{"ref":"StateServer.html#c:on_state_entry/3","title":"StateServer.on_state_entry/3","type":"callback","doc":"triggered on initialization or just prior to entering a state. If entering a state is done with a :goto statement or a :gen_statem state change, transition will be nil. Note that at this point the state change should not be cancelled. If you need to cancel a transition, use handle_transition/3 with the :cancel return value. response should be :noreply or a :noreply tuple with a restricted set of events which can be enqueued onto the events list: :update :internal :continue :event_timeout :state_timeout :timeout Like the other callbacks, you may call :defer here to defer to the state machines."},{"ref":"StateServer.html#reply/2","title":"StateServer.reply/2","type":"function","doc":"should be identical to GenServer.reply/2"},{"ref":"StateServer.html#start/3","title":"StateServer.start/3","type":"function","doc":""},{"ref":"StateServer.html#start_link/3","title":"StateServer.start_link/3","type":"function","doc":""},{"ref":"StateServer.html#c:terminate/3","title":"StateServer.terminate/3","type":"callback","doc":"triggered when the process is about to be terminated. See: c::gen_statem.terminate/3"},{"ref":"StateServer.html#t:defer_response/0","title":"StateServer.defer_response/0","type":"type","doc":"handler output when you want to defer to a state module"},{"ref":"StateServer.html#t:event/0","title":"StateServer.event/0","type":"type","doc":"events which can be put on the state machine&#39;s event queue. these are largely the same as t::gen_statem.event_type/0 but have been reformatted to be more user-friendly."},{"ref":"StateServer.html#t:noreply_response/0","title":"StateServer.noreply_response/0","type":"type","doc":"handler output when there isn&#39;t a response"},{"ref":"StateServer.html#t:on_state_entry_event/0","title":"StateServer.on_state_entry_event/0","type":"type","doc":"on_state_entry function outputs only a subset of the available handler responses should be queued from a triggered on_state_entry/3 event."},{"ref":"StateServer.html#t:reply_response/0","title":"StateServer.reply_response/0","type":"type","doc":"handler output when there&#39;s a response"},{"ref":"StateServer.html#t:stop_reply_response/0","title":"StateServer.stop_reply_response/0","type":"type","doc":"handler output for handle_call/4 which performs a stop action with a reply. If you would prefer stopping in an alternate form, you may enlist the help of reply/2."},{"ref":"StateServer.html#t:stop_response/0","title":"StateServer.stop_response/0","type":"type","doc":"handler output when the state machine should stop altogether. The value in new_data will be transferred as the data segment for terminate/3, so you may instrument important information there."},{"ref":"StateServer.html#t:timeout_payload/0","title":"StateServer.timeout_payload/0","type":"type","doc":""},{"ref":"StateServer.InvalidStateError.html","title":"StateServer.InvalidStateError","type":"exception","doc":"raised when you pass a funny atom to the {:goto, _} event"},{"ref":"StateServer.InvalidTransitionError.html","title":"StateServer.InvalidTransitionError","type":"exception","doc":"raised when you pass a funny atom to the {:transition, _} event"},{"ref":"StateServer.State.html","title":"StateServer.State","type":"behaviour","doc":"A behaviour that lets you organize code for your StateServer states. Organization When you define your StateServer, the StateServer module gives you the opportunity to define state modules. These are typically (but not necessarily) submodules scoped under the main StateServer module. In this way, your code for handling events can be neatly organized by state. In some (but not all) cases, this may be the most appropriate way to keep your state machine codebase sane. Defining the state module. the basic syntax for defining a state module is as follows: defstate MyModule, for: :state do # ... code goes here ... def handle_call(:increment, _from, data) do {:reply, :ok, update: data + 1} end end note that the callback directives defined in this module are identical to those of StateServer, except that they are missing the state argument. External state modules You might want to use an external module to handle event processing for one your state machine. Reasons might include: to enable code reuse between state machines if your codebase is getting too long and you would like to put state modules in different files. If you choose to do so, there is a short form defstate call, which is as follows: defstate ExternalModule, for: :state Be sure to mark your ExternalModule as having the StateServer.State behaviour. Precedence and Defer statements Note that handle_* functions written directly in the body of the StateServer take precedence over any functions written as a part of a state module. In the case where there are competing function calls, your handler functions written in the body of the StateServer may emit :defer as a result, which will punt the processing of the event to the state modules. # make sure query calls happen regardless of state def handle_call(:query, _from, _state, data) do {:reply, {state, data}} end # for all other call payloads, send to the state modules def handle_call(_, _, _, _) do :defer end defstate Start, for: :start do def handle_call(...) do... since this is a common pattern, we provide a defer macro which is equivalent to the above: # make sure query calls happen regardless of state def handle_call(:query, _from, _state, data) do {:reply, {state, data}} end # for all other call payloads, send to the state modules defer handle_call Important If you handle an event via any instance of a handler function block in the main StateServer module, and return a :reply or :noreply, it will not be handled by the State module, you must explicitly specify :defer to be handled by State modules. If there are no instances of the handler function, then handling will default to the State modules without using the defer macro. Defer with common events If you would like to perform analysis on the inbound data, generating events and defer to the individual states for further state-specific event processing, you may do so with the {:defer, events} result type. For example, the following code: # perform some common processing def handle_call({:query, payload}, _from, _state, data) do common_events = generate_common_events_from(payload) {:defer, common_events} end defstate Start, for: :start do def handle_call({:query, payload}, _from, data) do # ...some code here... {:reply, result, start_events} end end Will result in the event stream common_events ++ start_events emitted when {:query, payload} is called to the state server, with the following exception: If you have an {:update, &lt;new_data&gt;} in the first position, or in the second position with a {:goto, &lt;state&gt;}, or {:transition, &lt;state&gt;} in the first position, the update event will be reflected in the deferred state machine call. Termination rules If a State module implements the terminate/2 callback, then it will be called on termination. If it does not, termination will follow the parent StateServer&#39;s StateServer.terminate/3 if it exists. Otherwise, no action will be taken on terminate. Example The following code should produce a &quot;light switch&quot; state server that announces when it&#39;s been flipped. defmodule SwitchWithStates do @doc &quot;&quot;&quot; implements a light switch as a state server. In data, it keeps a count of how many times the state of the light switch has changed. On transition, it sends to standard error a comment that it has been flipped. Note that the implementations are different between the two states. &quot;&quot;&quot; use StateServer, off: [flip: :on], on: [flip: :off] @type data :: non_neg_integer def start_link, do: StateServer.start_link(__MODULE__, :ok) @impl true def init(:ok), do: {:ok, 0} def flip(srv), do: StateServer.call(srv, :flip) def query(srv), do: StateServer.call(srv, :query) @impl true def handle_call(:flip, _from, _state, _count) do {:reply, :ok, transition: :flip} end defer handle_call # we must defer the handle_call statement because there are both shared and # individual implementation of handle_call features. defstate Off, for: :off do @impl true def handle_transition(:flip, count) do IO.puts(:stderr, &quot;switch \#{inspect self()} flipped on, \#{count} times turned on&quot;) {:noreply, update: count + 1} end @impl true def handle_call(:query, _from, _count) do {:reply, &quot;state is off&quot;} end end defstate On, for: :on do @impl true def handle_transition(:flip, count) do IO.puts(:stderr, &quot;switch \#{inspect self()} flipped off, \#{count} times turned on&quot;) :noreply end @impl true def handle_call(:query, _from, _count) do {:reply, &quot;state is on&quot;} end end end"},{"ref":"StateServer.State.html#c:handle_call/3","title":"StateServer.State.handle_call/3","type":"callback","doc":""},{"ref":"StateServer.State.html#c:handle_cast/2","title":"StateServer.State.handle_cast/2","type":"callback","doc":""},{"ref":"StateServer.State.html#c:handle_continue/2","title":"StateServer.State.handle_continue/2","type":"callback","doc":""},{"ref":"StateServer.State.html#c:handle_info/2","title":"StateServer.State.handle_info/2","type":"callback","doc":""},{"ref":"StateServer.State.html#c:handle_internal/2","title":"StateServer.State.handle_internal/2","type":"callback","doc":""},{"ref":"StateServer.State.html#c:handle_timeout/2","title":"StateServer.State.handle_timeout/2","type":"callback","doc":""},{"ref":"StateServer.State.html#c:handle_transition/2","title":"StateServer.State.handle_transition/2","type":"callback","doc":""},{"ref":"StateServer.State.html#c:on_state_entry/2","title":"StateServer.State.on_state_entry/2","type":"callback","doc":""},{"ref":"StateServer.State.html#c:terminate/2","title":"StateServer.State.terminate/2","type":"callback","doc":""},{"ref":"StateServer.StateGraph.html","title":"StateServer.StateGraph","type":"module","doc":"tools for dealing with stategraphs. State graphs take the form of a keyword list of keyword lists, wherein the outer list is a comprehensive list of the states, and the inner lists are keywords list of all the transitions, with the values of the keyword list as the destination states."},{"ref":"StateServer.StateGraph.html#all_transitions/1","title":"StateServer.StateGraph.all_transitions/1","type":"function","doc":"lists all state/transition pairs. Used to generate the StateServer.is_transition/2 guard. Example iex&gt; StateServer.StateGraph.all_transitions([start: [t1: :state1, t2: :state2], state1: [], state2: [t2: :state1]]) [start: :t1, start: :t2, state2: :t2]"},{"ref":"StateServer.StateGraph.html#atoms_to_typelist/1","title":"StateServer.StateGraph.atoms_to_typelist/1","type":"function","doc":"converts a list of atoms to a type which is the union of the atom literals"},{"ref":"StateServer.StateGraph.html#edges/1","title":"StateServer.StateGraph.edges/1","type":"function","doc":"outputs a list of edges of the graph. Used to generate the StateServer.is_transition/3 guard. iex&gt; StateServer.StateGraph.edges(start: [t1: :state1, t2: :state2], state1: [t3: :start], state2: []) [start: {:t1, :state1}, start: {:t2, :state2}, state1: {:t3, :start}]"},{"ref":"StateServer.StateGraph.html#start/1","title":"StateServer.StateGraph.start/1","type":"function","doc":"returns the starting state from the state graph. iex&gt; StateServer.StateGraph.start([start: [t1: :state1], state1: [t2: :state2], state2: []]) :start"},{"ref":"StateServer.StateGraph.html#states/1","title":"StateServer.StateGraph.states/1","type":"function","doc":"lists all states in the state graph. The first element in this list will always be the initial state. Example iex&gt; StateServer.StateGraph.states([start: [t1: :state1], state1: [t2: :state2], state2: [t2: :state2]]) [:start, :state1, :state2]"},{"ref":"StateServer.StateGraph.html#terminal_states/1","title":"StateServer.StateGraph.terminal_states/1","type":"function","doc":"outputs a list of terminal states of the graph. Used to generate the StateServer.is_terminal/1 guard. iex&gt; StateServer.StateGraph.terminal_states(start: [t1: :state1, t2: :state2], state1: [], state2: []) [:state1, :state2]"},{"ref":"StateServer.StateGraph.html#terminal_transitions/1","title":"StateServer.StateGraph.terminal_transitions/1","type":"function","doc":"outputs a list of terminal {state, transition} tuples of the graph. Used to generate the StateServer.is_terminal/2 guard. iex&gt; StateServer.StateGraph.terminal_transitions(start: [t1: :state1, t2: :state2], state1: [], state2: []) [start: :t1, start: :t2]"},{"ref":"StateServer.StateGraph.html#transition/3","title":"StateServer.StateGraph.transition/3","type":"function","doc":"outputs the destination state given a source state and a transition. Example iex&gt; StateServer.StateGraph.transition([start: [t1: :state1, t2: :state2], state1: [], state2: []], :start, :t1) :state1"},{"ref":"StateServer.StateGraph.html#transitions/1","title":"StateServer.StateGraph.transitions/1","type":"function","doc":"lists all transitions in the state graph. Example iex&gt; StateServer.StateGraph.transitions([start: [t1: :state1], state1: [t2: :state2], state2: [t2: :state2]]) [:t1, :t2]"},{"ref":"StateServer.StateGraph.html#transitions/2","title":"StateServer.StateGraph.transitions/2","type":"function","doc":"lists all transitions emanating from a given state. Example iex&gt; StateServer.StateGraph.transitions([start: [t1: :state1, t2: :state2], state1: [], state2: []], :start) [:t1, :t2]"},{"ref":"StateServer.StateGraph.html#valid?/1","title":"StateServer.StateGraph.valid?/1","type":"function","doc":"checks the validity of the state graph. Should only be called when we build the state graph. A state graph is valid if and only if all of the following are true: the graph is a keyword of keywords. there is at least one state. there are no duplicate state definitions. there are no duplicate transition definitions emanating from any given state. all of the destinations of the transitions are states."},{"ref":"StateServer.StateGraph.html#t:t/0","title":"StateServer.StateGraph.t/0","type":"type","doc":""},{"ref":"readme.html","title":"StateServer","type":"extras","doc":"StateServer"},{"ref":"readme.html#opinionated-gen_statem-shim-for-elixir","title":"StateServer - Opinionated :gen_statem shim for Elixir","type":"extras","doc":"A foolish consistency is the hobgoblin of little minds, adored by little statemen and philosophers and divines. With consistency a great soul has simply nothing to do. -- Ralph Waldo Emerson The unusual callback pattern of :gen_statem exists to allow code organization which we have better ways of achieving in Elixir (versus erlang). On the other hand we want to make sure users are using the canonical :gen_statem to leverage and prove out its battle-testedness This library makes :gen_statem more consistent with how Elixir architects its GenServers. There are three major objectives: Fanout the callback handling Unify callback return type with that of GenServer, and sanitize Enforce the use of a programmer-defined state graph."},{"ref":"readme.html#installation","title":"StateServer - Installation","type":"extras","doc":"The package can be installed by adding state_server to your list of dependencies in mix.exs: def deps do [ {:state_server, &quot;~&gt; 0.4.7&quot;} ] end The docs can be found at https://hexdocs.pm/state_server."},{"ref":"readme.html#example","title":"StateServer - Example","type":"extras","doc":"defmodule Demo do use StateServer, on: [flip: :off], off: [flip: :on] def start_link(_), do: StateServer.start_link(__MODULE__, [], name: Demo) def init(state), do: {:ok, []} def handle_call(:flip, _from, state, data) do {:reply, data, transition: :flip, update: [state | data], timeout: {:foo, 100}} end def handle_transition(start, tr, data) do IO.puts(&quot;transitioned from \#{start} through \#{tr}&quot;) :noreply end def handle_timeout(_, _, _) do IO.puts(&quot;timed out!&quot;) :noreply end end iex(2)&gt; Demo.start_link(:ok) {:ok, #PID&lt;0.230.0&gt;} iex(3)&gt; GenServer.call(Demo, :flip) transitioned from on through flip [] timed out! iex(4)&gt; GenServer.call(Demo, :flip) transitioned from off through flip [:on] timed out! iex(5)&gt; GenServer.call(Demo, :flip) transitioned from on through flip [:off, :on] timed out!"}]